/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL1_Expand_Motifs  (Level 1 library routine)                       */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    22 JUL 94   Version 1.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure transforms a list of motifs as loaded from a Transcription    */
/* Factor Database file(s) into "internal" format, more convenient for       */
/* further processing; returns the "standard" error structure, NULL if there */
/* were no errors                                                            */
/*                                                                           */
/* NOTE: The procedure is "destructive" for the received list of motifs:     */
/*       after they are copied and rearranged, the cells are destroyed       */
/*       (returned to the system) and the pointer to the head of the list is */
/*       set to NULL                                                         */
/*                                                                           */
/*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl1.h"



errind NTL1_EM_Assemble_Error (int severity, int code,
                               char *comment, int description);

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_Expand_Motifs                                             */
/*                                                                           */
/* Procedure to execute expansion; returns the error structure (or NULL)     */

errind NTL1_Expand_Motifs (tfd_ptr *motif_list, char **codes,
                           char **complements, motif_ptr *expanded)
{
 tfd_ptr current, old_one; motif_ptr new_motif; int index; char *scan;

 *expanded = NULL;                /* Initialize the expanded list to "empty" */

 /* Loop through the received list of motifs - process each record within    */

 current = *motif_list; while (current != NULL) {
  old_one = current;

  new_motif = (motif_ptr) NTL0_ckalloc (sizeof (Motif_Struct));

  new_motif -> name = NULL;
  new_motif -> originating = current -> file_name;
  new_motif -> bind_factor = current -> bind_factor;
  new_motif -> site_name = current -> site_name;

  if (current -> seq_length != strlen (current -> sequence))
   return NTL1_EM_Assemble_Error (USER_ERROR, ERR_BAD_STRUCTURE,
                                  "Bad motif length", 1);
  else new_motif -> length = current -> seq_length;

  new_motif -> text = current -> sequence;

  /* Now do the expansion and conversion of motif text with respect to codes */

  new_motif -> coded = (long int *)
                  NTL0_ckalloc ((new_motif -> length) * sizeof (long int));
  for (index = 0; index < new_motif -> length; index++) {
   (new_motif -> coded) [index] = 0;
   scan = codes [(new_motif -> text) [index] - 'A'];
   while (*scan != '\0') {
    (new_motif -> coded) [index] = ((new_motif -> coded) [index]) |
                                    (((long int) 1) << ((int) (*scan - 'A')));
    scan++;
   }
  }

  new_motif -> inverted = (long int *)
                  NTL0_ckalloc ((new_motif -> length) * sizeof (long int));
  for (index = 0; index < new_motif -> length; index++) {
   (new_motif -> inverted) [(new_motif -> length) - index - 1] = 0;
   scan = complements [(new_motif -> text) [index] - 'A'];
   while (*scan != '\0') {
    (new_motif -> inverted) [(new_motif -> length) - index - 1] =
     ((new_motif -> inverted) [(new_motif -> length) - index - 1]) |
                                      (((long int) 1) << ((int) (*scan - 'A')));
    scan++;
   }
  }

  new_motif -> consensus = FALSE;
  new_motif -> global = new_motif -> local = 0.0;
  new_motif -> outline = 0;
  new_motif -> style = NULL;
  new_motif -> label_pos = DEFAULT_POSITION;
  new_motif -> label_place = DEFAULT_LABEL;
  new_motif -> label = NULL;
  new_motif -> ukind = 0;
  new_motif -> at_seq = NULL;

  new_motif -> reference = current -> reference;

  new_motif -> next = *expanded; *expanded = new_motif;

  current = current -> next;
  free (old_one);
 }
 *motif_list = NULL; return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_EM_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL1_EM_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + strlen ("_Expand_Motifs: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Expand_Motifs: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

