/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL2_Get_Alphabet  (Level 2 library routine)                        */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    06 MAY 96   Version 1.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure fills in the provided arrays with alphabet information: row   */
/* index for matrices and letter relative frequences for all permitted       */
/* letters.                                                                  */
/*   Procedure receives two alphabet-size (26) arrays, integer and double    */
/* float, and fills them with the right contents; returns the error          */
/* structure, or NULL if everything was OK                                   */
/*                                                                           */
/*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl2.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limit for the length of an error message that can be assembled            */

#define ERR_MSGLIMIT             128

/* Indices into row array of the matrix for permitted alphabet letters       */

#define INDEX_A                    0
#define INDEX_C                    1
#define INDEX_G                    2
#define INDEX_T                    3

/* Relative frequences of the permitted alphabet letters                     */

#define RELATIVE_FREQUENCY_A     0.3
#define RELATIVE_FREQUENCY_C     0.2
#define RELATIVE_FREQUENCY_G     0.2
#define RELATIVE_FREQUENCY_T     0.3


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL2_GA_Assemble_Error (int severity, int code,
                               char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_Get_Alphabet                                              */
/*                                                                           */
/* Main (interface) procedure of this unit                                   */

errind NTL2_Get_Alphabet (int *Indices, double *Frequences)
{
 int index;

 for (index = 0; index < 26; index++) Indices [index] = -1;
 for (index = 0; index < 26; index++) Frequences [index] = 0.0;

 Indices [(int) 'A' - (int) 'A'] = INDEX_A;
 Indices [(int) 'C' - (int) 'A'] = INDEX_C;
 Indices [(int) 'G' - (int) 'A'] = INDEX_G;
 Indices [(int) 'T' - (int) 'A'] = INDEX_T;

 Frequences [(int) 'A' - (int) 'A'] = RELATIVE_FREQUENCY_A;
 Frequences [(int) 'C' - (int) 'A'] = RELATIVE_FREQUENCY_C;
 Frequences [(int) 'G' - (int) 'A'] = RELATIVE_FREQUENCY_G;
 Frequences [(int) 'T' - (int) 'A'] = RELATIVE_FREQUENCY_T;
 
 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_GA_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL2_GA_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + strlen ("_Get_Alphabet: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Get_Alphabet: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

