/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL2_Read_TRANSFAC_References  (Level 2 library routine)            */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    10 APR 97   Version 1.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure reads the specified references from the TRANSFAC (Wingender)  */
/* Database references file, in the "plain text" format, into a vector of    */
/* string lists with full reference descriptions (string lists are used in   */
/* order to accomodate multiple citations per reference code); due to the    */
/* size of the references file only the specified references are read.       */
/*   Procedure receives the name of the file containing TRANSFAC references  */
/* catalogue, number of requested references and vector of character strings */
/* with reference codes; fills in the full reference information into        */
/* another vector of string lists. Procedure returns the error structure,    */
/* NULL if everything was OK                                                 */
/*                                                                           */
/*                                                                           */
/* Expected format of the TRANSFAC (Wingender) database references file:     */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl2.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limit for the length of an error message that can be assembled            */

#define ERR_MSGLIMIT             128

/* Information is read from database files by supplementary buffer of size:  */

#define IN_LIMIT                 200

/* Delimiting positions for various blocks of information within a line      */

#define REF_CODE_START           1
#define REF_CODE_STOP            8

#define SITE_INFO_START          9
#define SITE_INFO_STOP           45

#define REFERENCE_START          46
#define REFERENCE_STOP           180


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


void NTL2_RW_Sort_Selections (int number, char **selections);
errind NTL2_RW_Read_References (FILE **ref_file, int number,
				char **selections, strlist_ptr *details);
errind NTL2_RW_Assemble_Error (int severity, int code,
                               char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char In_Buff [IN_LIMIT]; /* Global input line buffer, for convenience */

static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_Read_TRANSFAC_References                                  */
/*                                                                           */
/* Main (interface) procedure of this unit                                   */

errind NTL2_Read_TRANSFAC_References (char *file_name, int number,
                                      char **selections, strlist_ptr *details)
{
 FILE *ref_file; int index; errind erret; strlist_ptr exp_scan, exp_rec;

 /* Check the legality of the received information and open the file         */

 if (number <= 0)
  return NTL2_RW_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 "Illegal number of references requested", 1);
                                 
 if ((ref_file = fopen (file_name, "r")) == NULL) {       /* Error condition */
  sprintf (Error_Message, "File <%s> does not exist", file_name);
  return NTL2_RW_Assemble_Error (USER_ERROR, ERR_NO_FILE, Error_Message, 0);
 }
 else {                           /* There is a file with the specified name */
  /* Initialize the expanded details vector                                  */

  for (index = 0; index < number; index++) details [index] = NULL;

  /* Sort the received list of selections in alphabetical order              */

  NTL2_RW_Sort_Selections (number, selections);

  /* Now load the references from the file and close the file when done      */

  erret = NTL2_RW_Read_References (&ref_file, number, selections, details);
  if ((erret != NULL) && (erret -> kind != WARNING)) {      /* Serious error */
   /* Release all the data allocated so far in the process                   */

   for (index = 0; index < number; index++) if (details [index] != NULL) {
    exp_scan = details [index]; while (exp_scan != NULL) {
     exp_rec = exp_scan; exp_scan = exp_scan -> next;
     if (exp_rec -> string != NULL) free (exp_rec -> string); free (exp_rec);
    }
    free (details [index]); details [index] = NULL;
   }
  }
  fclose (ref_file); return erret;
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_RW_Sort_Selections                                        */
/*                                                                           */
/* Procedure sorts the received vector of strings in the alphabetical order  */

void NTL2_RW_Sort_Selections (int number, char **selections)
{
 int last, current; char *temp;

 for (last = number - 1; last >= 1; last--) {
  for (current = 0; current < last; current++) {
   if (strcmp (selections [current], selections [current + 1]) > 0) {
    temp = selections [current];
    selections [current] = selections [current + 1];
    selections [current + 1] = temp;
   }
  }
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_RW_Read_References                                        */
/*                                                                           */
/* Top level procedure for reading the TRANSFAC (Wingender) database         */
/*   references file contents (only the records from the provided list,      */
/*   though); loops for each reference (line) defined in the file; receives  */
/*   the file pointer, number of records to read and vector of reference     */
/*   codes to search for; fills in the vector of expanded references;        */
/*   returns the assembled error record, NULL if there were no errors in     */
/*   loading                                                                 */

errind NTL2_RW_Read_References (FILE **ref_file, int number,
                                char **selections, strlist_ptr *details)
{
 int index, space_count; strlist_ptr new_cell, cell_scan;
 char save_char, *check_string, *prev_str, *scan, *expanded;

 index = 0; prev_str = NULL; check_string = NULL;
 while ((index < number) && (fgets (In_Buff, IN_LIMIT, *ref_file) != NULL) &&
        (In_Buff [0] != '\0')) {
  scan = In_Buff; while ((*scan == ' ') || (*scan == '\t')) scan++;
  if (*scan == '\0')
   return NTL2_RW_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                                  "Unexpected end of string in line", 0);
  else if (*scan != '\n') {
   if (strlen (In_Buff) < REFERENCE_START)
    return NTL2_RW_Assemble_Error (USER_ERROR, ERR_FILE_FORMAT,
                       "Illegal line length in TRANSFAC reference file", 0);
                       
   /* Locate the abbreviation for the reference described in the line        */
   
   space_count = REF_CODE_STOP - 1;
   while (((In_Buff [space_count] == ' ') || (In_Buff [space_count] == '\t') ||
           (In_Buff [space_count] == '\n')) && (space_count >= REF_CODE_START))
    space_count--;
   if ((In_Buff [space_count] != ' ') && (In_Buff [space_count] != '\t') &&
       (In_Buff [space_count] != '\n') && (In_Buff [space_count] != '\0'))
    space_count++;
   save_char = In_Buff [space_count]; In_Buff [space_count] = '\0';
   check_string = NTL0_strsave (&(In_Buff [REF_CODE_START - 1]));
   In_Buff [space_count] = save_char;

   if (prev_str != NULL) {       /* Examine end of current list of citations */
    if (strcmp (prev_str, check_string)) {
     free (prev_str); prev_str = NULL; index++;
    }
   }
   /* Check now whether there is a match between the next code and this line */
   
   if ((index < number) && (!strcmp (check_string, selections [index]))) {
    space_count = strlen (In_Buff) - 1;
    while (((In_Buff [space_count] == ' ') ||
            (In_Buff [space_count] == '\t') ||
            (In_Buff [space_count] == '\n')) &&
           (space_count >= REFERENCE_START)) space_count--;
    if ((In_Buff [space_count] != ' ') && (In_Buff [space_count] != '\t') &&
        (In_Buff [space_count] != '\n'))
     space_count++;
    save_char = In_Buff [space_count]; In_Buff [space_count] = '\0';
    expanded = NTL0_strsave (&(In_Buff [REFERENCE_START - 1]));
    In_Buff [space_count] = save_char;

    new_cell = (strlist_ptr) NTL0_ckalloc (sizeof (StrList_Struct));
    new_cell -> string = expanded;
    new_cell -> next = NULL;

    if (prev_str == NULL) {
     details [index] = new_cell; prev_str = check_string;
    }
    else {
     cell_scan = details [index];
     while (cell_scan -> next != NULL) cell_scan = cell_scan -> next;
     cell_scan -> next = new_cell;
    }
   }
   if (prev_str == NULL) { free (check_string); check_string = NULL; }
  }
  In_Buff [0] = '\0';
 }
 if (strlen (In_Buff) > 0)
  return NTL2_RW_Assemble_Error (SYSTEM_ERROR, ERR_BAD_VALUE,
                                 "Unprocessed string in the input buffer", 0);
 else if (index < number - 1)
  return NTL2_RW_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 "Reference code not found", 0);
 else if ((prev_str == NULL) && (index < number))
  return NTL2_RW_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 "Reference code not found", 0);
 else {
  if (prev_str != NULL) free (prev_str);
  return NULL;
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL2_RW_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL2_RW_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
             (strlen (comment) + strlen ("_Read_TRANSFAC_References: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Read_TRANSFAC_References: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

